import pandas as pd
import os
from fdn_kg_loader.model import Brand, Product, Category, CategoryTree, CategorizedRel, ProducedRel
from fdn_kg_loader.store import Store
from fdn_kg_loader.filter import *
from glob import glob
import os
import pandas as pd


# TODO : add configure command

# TODO : automatic data splitting for brand
# TODO : automatic data splitting for product
# TODO : automatic data splitting for category
# TODO : automatic data splitting for distribution channel
# TODO : automatic splitting for relation
# TODO : automatic tagger for shade instead of variant by looking up to <has_shade> column


def load_file_chunk(base_path, path_pattern):
    # CURDIR = os.path.dirname(os.path.abspath('.'))
    # FDIR = os.path.join(CURDIR, 'fdn-annotator/shapeshift/reviews/*')
    # FDIR = os.path.join(CURDIR,'fdn-annotator/*reviews.csv')
    FDIR = os.path.join(base_path,path_pattern)
    files = [file for file in glob(FDIR)]

    N = len(files)

    df=[]
    for i in range(N):
        df0 = pd.read_csv(files[i])
        df.append(df0)
    df = pd.concat(df,sort=False)

    return df


def load_brand(df, xstore):
    df = df[['gid','name']]
    df = df.drop_duplicates()
    print(df)
    brands = df.to_dict(orient='record')
    for brand in brands:
        o = Brand()
        o.from_dict(brand)
        # xstore.update(o,update_by=["gid"])
        xstore.save(o)
    xstore.flush()



    #significant part start from here
    
        
def _load_undup_product(df,xstore):
    
    undup_df = df[df['name'].duplicated()==False]
#     undup_df = undup_df.head()
    prods = undup_df.to_dict(orient='record')
    try:
        for prod in prods:
            o = Product()
            o.from_dict(prod)
            print(o.__dict__)
            xstore.save(o)    
        xstore.flush()
    except Exception as e:
                print("insert failed")
                print(e)
                

def _load_dup_product(df,xstore):
    dup_df = df[df['name'].duplicated()==True]
    prods = dup_df.to_dict(orient='record')
    try:
        for prod in prods:
            o = Product()
            o.from_dict(prod)
            print(o.__dict__)
            xstore.update(o,update_by=['name'])
        xstore.flush()
    except Exception as e:
        print("already existed, skipping update")
        print(e)


def load_product(df,xstore):
    
    df = df[['gid','name','variant']]
    df = df.fillna('')

    _load_undup_product(df,xstore)
    _load_dup_product(df,xstore)

def _load_category_node(df,xstore):
    print("loading category node")
    df = df[['cat_id','cat_item']]
    df = df.rename(columns={'cat_id':'gid','cat_item':'name'})
    print(df.head())
    categories = df.to_dict(orient='record')
    for c in categories:
        o = Category()
        o.from_dict(c)
        print(o.__dict__)
        try:
        #    xstore.update(o,update_by=['gid'])
            xstore.save(o)
        except Exception as e:
           print(e)
    xstore.flush()
        

def _load_category_tree(df,xstore):
    print("loading category tree")
    df = df.rename(columns={'cat_id':'current','cat_parent':'parent'})
    print(df.head())
    category_tree = df.to_dict(orient='record')
    for r in category_tree:
        t = CategoryTree().relatesObj(
                entity='category',identifier='gid',
                value=r['current'], plays='current').withObj(
                    entity='category',identifier='gid',
                    value=r['parent'], plays='parent')
        # print(t.__dict__)
        xstore.relates(t,filter=lambda l,r: l['value'] != 0 and r['value'] != 0 )
    xstore.flush()


def load_category(df,xstore):
    _load_category_node(df,xstore)

def build_category_tree(df,xstore):
    _load_category_tree(df,xstore)


def relate_category_product(df, xstore):
    rel_list = df.to_dict(orient='record')
    for r in rel_list:
        rel = CategorizedRel().relatesObj(entity='product',identifier='gid',
                value=r['product_gid'],plays='categorized-as').withObj(entity='category',
                        identifier='gid',value=r['category_gid'],plays='container-of')
        xstore.relates(rel)
    xstore.flush()


    
def relate_product_brand(df2, xstore):
    rel_list = df2.to_dict(orient='record')
    for r in rel_list:
        rel = ProducedRel().relatesObj(entity='product',identifier='gid',
                value=r['product_gid'],plays='produce').withObj(entity='brand',
                        identifier='gid',value=r['brand_gid'],plays='produced-by')
        xstore.relates(rel)
    xstore.flush()






from ..model import Product
from ..store import Store
import unittest

class TestProduct(unittest.TestCase):
    def setUp(self):
        self.host = "127.0.0.1"
        self.keyspace = "schema_check"
    def tearDown(self):
        pass

    def test_insert(self):
        o = {"name":"Test Product","gid":"0001"}        
        p = Product()
        p.from_dict(o)
        self.assertIsNotNone(p)
        store = Store(host=self.host, keyspace = self.keyspace)
        store.save(p)
        store.flush()

        r = store.query('match $o isa product, has gid "0001"; get;')
        result = [x for x in r]
        print(result)
        self.assertTrue(len(result) > 0)

    def test_double_insert(self):
        pass




from grakn.client import GraknClient
import copy, re
import os
from multiprocessing import Process

def process(L,host='localhost',port=48555,keyspace='grakn'):
    with GraknClient(uri="{}:{}".format(host, port)) as client:
        with client.session(keyspace=keyspace) as sess:
            with sess.transaction().write() as wtx:
                for s in L:
                    # s = 'insert $o isa number, has value "p-{}";'.format(i)
                    # print("PID : {} , Q : {}".format(os.getpid(), s))
                    # raise Exception("test this exception")
                    wtx.query(s)

                wtx.commit()
                print("committing from pid {}".format(os.getpid()))
class Store:
    
    session = None
    tx = None

    def __init__(self, host='localhost', port=48555, keyspace='grakn'):
        self.HOST = host
        self.PORT = port
        self.KEYSPACE = keyspace
        self._client = None
        self.bulk_query = []

        # self._client = self._make_client()
        self._session = GraknClient(uri="{}:{}".format(self.HOST, self.PORT)).session(keyspace=self.KEYSPACE)

    def _close(self):
        if self._session is not None:
            self._session.close()
            # self._client.close()
    
    def __del__(self):
        self._close()
        
    def exec(self, q):
        with self._session.transaction().write() as tx:
            #rqit is read query iterator
            rqit = tx.query(q)
            # concepts = rqit.collect_concepts()
            # id = concepts[0].id
            tx.commit()

    def query(self, q):
        
        with self._session.transaction().read() as tx:
            #rqit is read query iterator
            rqit = tx.query(q)
            rs = rqit.collect_concepts()
            for r in rs:
                yield r
                    
    
    def _make_client(self):
        return GraknClient(uri="{}:{}".format(self.HOST, self.PORT))

    def relates(self, rel, filter=None):
        def _build_query(props,relation_name):
            Left, Right = props['relates']

            query = '''
                match 
                    $l isa {}, has {} "{}";
                    $r isa {}, has {} "{}";
                insert 
                    $R({}: $l , {}:$r) isa {};
            '''.format(Left['entity'],Left['identifier'],Left['value'],
                        Right['entity'],Right['identifier'],Right['value'],
                        Left['as'],Right['as'],relation_name)
            # print(query)
            return re.sub(r'\s+',' ',query)

        relation_name = getattr(rel.__class__,'relation_name')
        props = rel.__dict__
        query = _build_query(props, relation_name)
        Left, Right = props['relates']
        if filter is None:
            self.bulk_query.append(query)
        elif filter(Left, Right):
            self.bulk_query.append(query)
        
    def save(self, o):
        '''
            save here actually upsert. :
                two operation done : 1. delete, 2. insert
        '''
        # with self._make_client() as client:
        #     with client.session(keyspace=self.KEYSPACE) as session:
        #         with session.transaction().write() as wtx:

        entity_name = getattr(o.__class__,'entity_name')
        props = o.__dict__
        fields = ', '.join([ 'has {} "{}"'.format(key, props[key]) for key in props if props[key] != ''])
        # query = 'match $o isa {}, {}; delete $o; insert $o isa {}, {}; '.format(entity_name, fields, entity_name,fields)
        
        delq = 'match $o isa {}, {}; delete $o;'.format(entity_name, fields)
        insq = 'insert $o isa {}, {}; '.format(entity_name,fields)
        self.bulk_query.append(delq)
        self.bulk_query.append(insq)

    def flush(self):
        n = 10
        ntask = len(self.bulk_query)
        groups = [ self.bulk_query[i:i+n] for i in range(0, ntask, n) ]

        pn = 0
        ps = []
        for t in groups:
            p = Process(target=process, args=(t,self.HOST,self.PORT,self.KEYSPACE,))
            ps.append(p)
            ps[pn].start()
            pn += 1

        for i in range(pn) :
            ps[i].join()
        
        del self.bulk_query
        self.bulk_query = []

    def update(self, o, update_by):
        
        entity_name = getattr(o.__class__,'entity_name')
        props = o.__dict__
        crit_prop = copy.deepcopy(o.__dict__)
        criteria = [(key, crit_prop[key]) for key in update_by]
        print("props: {}, crit : {}, searchc: {}".format(props, crit_prop, criteria))
        # del props['name']
        # del props['gid']
        print("props: {}, crit : {}, searchc: {}".format(props, crit_prop, criteria))

        fields = ', '.join([ 'has {} "{}"'.format(key, props[key]) for key in props if props[key] != '' and key != 'gid' ])
        criteria = ', '.join(['has {} "{}"'.format(k, v) for k, v in criteria])
        if len(fields) > 0:
            query = 'match $x isa {}, {} ; insert $x {}; '.format(entity_name,criteria,fields)
            self.bulk_query.append(query)


import unittest
from fdn_kg_loader.store import Store
from fdn_kg_loader.model import Brand
'''
    for test : 

        load fdn-annotator's env
        python -m unittest 
'''
class TestBrandInsert(unittest.TestCase):
    def setUp(self):

        print("setup")
        self.host = "127.0.0.1"
        self.keyspace = "schema_check"
    def tearDown(self):
        store = Store(host=self.host, keyspace=self.keyspace)
        print("check brand with gid 9999 : (should return one)")
        
        objs = store.query('match $o isa brand, has gid "9999"; get;')
        print([x for x in objs])
        print("check done.")
        print("deleting test data")
        store.exec('match $o isa brand, has gid "9999"; delete $o;')
        
    def test_insert(self):

        b = Brand()
        b.from_dict({
            'gid':'9999',
            'name':'Test Object'
        })
        store = Store(host=self.host,keyspace=self.keyspace)
        store.save(b)
        store.flush()
        # store.close()

    def test_duplicate(self):
        pass
        #save save flush
        # b = Brand()
        # b.from_dict({
        #     'gid':'9999',
        #     'name':'Test Object'
        # })
        # store = Store(host="127.0.0.1",keyspace="schema_check")
        # store.save(b)
        # store.save(b)
        # store.flush()

    def test_duplicate_and_reconnect(self):
        #save flush - save flush
        pass
        # b = Brand()
        # b.from_dict({
        #     'gid':'9999',
        #     'name':'Test Object'
        # })
        # store = Store(host="127.0.0.1",keyspace="schema_check")
        # store.save(b)
        # store.flush()
        # store.save(b)
        # store.flush()

    def test_different_connectoin(self):
        b = Brand()
        b.from_dict({
            'gid':'9999',
            'name':'Test Object'
        })
        store = Store(host="127.0.0.1",keyspace="schema_check")
        store2 = Store(host="127.0.0.1",keyspace="schema_check")
        store.update(b,update_by=['gid'])
        store.flush()
        store2.update(b,update_by=['gid'])
        store2.flush()

if __name__ == "__main__":
    unittest.main()


def filter_product_category_relation(df):
    '''
        should pass df
    '''
    prodcat_df = df[['prod_id','cat_id']]
    prodcat_df = prodcat_df.rename(columns={'prod_id':'product_gid','cat_id':'category_gid'})
    prodcat_df = prodcat_df.drop_duplicates()
    return prodcat_df

def filter_product_brand_relation(df):
    prodbrand_df = df[['prod_id','brands_id']]
    prodbrand_df = prodbrand_df.rename(columns={'prod_id':'product_gid','brands_id':'brand_gid'})
    return prodbrand_df

def filter_category_tree_relation(df):
    categtree_df = df[['cat_id','cat_parent']]
    categtree_df = categtree_df.drop_duplicates()
    return categtree_df

def filter_category_entity(df):
    cat_df = df[['cat_id','cat_parent','cat_item','cat_status','cat_has_shade']]
    cat_df = cat_df.drop_duplicates()
    return cat_df

def filter_brand_entity(df):
    brand_df = df[['brands_id','brands_item','brands_slug','brands_desc']]
    brand_df = brand_df.rename(columns={'brands_id':'gid','brands_item':'name','brands_slug':'slug','brands_desc':'desc'})
    brand_df = brand_df[['gid','name','slug']]
    brand_df = brand_df.drop_duplicates()
    return brand_df

def filter_product_entity(df):
    product_df = df
    product_df = product_df[product_df['review_deletestatus'] != 'y']
    product_df = product_df.fillna('').sort_values(by=['prod_item']).drop_duplicates()
    product_df = product_df.rename(columns={'prod_id':'gid','brands_id':'brand_gid','prod_item':'name','prod_slug':'slug','prod_type':'variant'})
    product_df = product_df[['gid','brand_gid','name','slug','variant']]
    return product_df


import setuptools

setuptools.setup(
    name='fdn-kg-loader',
    version='0.14.2',
    packages=setuptools.find_packages(),
    license='Restricted License, FDN license',
    long_description=open('README.md').read(),
    entry_points={
        'console_scripts': ['fdn_kg_loader=fdn_kg_loader:main']
    },
    install_requires=[
        'pandas',
        'grpcio',
        'protobuf',
        'grakn',
        'grakn-client',
        'click',
        'six'
    ]
)
